#define _GNU_SOURCE
#include <stdio.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>

#include "mtree.h"
#include "hashfile.h"

#define FILE_BUFFER			0x1000

static char hex[16] =
	"0123456789abcdef";

#define HASHFile(F, CTX, Init, Update, Final, N) \
char *F(const char *pathname, char *output) \
{ \
	CTX c; \
	unsigned char binary[N]; \
	struct stat st; \
	int fd, n, i; \
	char *buffer, *p; \
\
	if (stat(pathname, &st)) return NULL; \
	if (!S_ISREG(st.st_mode)) { \
		errno = EIO; \
		return NULL; \
	} \
\
	if ((fd = open(pathname, MTREE_O_FLAGS)) < 0) \
		return NULL; \
\
	if (fstat(fd, &st)) { \
		close(fd); \
		return NULL; \
	} \
	if (!S_ISREG(st.st_mode)) { \
		close(fd); \
		errno = EIO; \
		return NULL; \
	} \
\
	if (!(buffer = malloc(FILE_BUFFER))) { \
		close(fd); \
		errno = ENOMEM; \
		return NULL; \
	} \
\
	Init(&c); \
	while ((n = read(fd, buffer, FILE_BUFFER)) > 0) \
		Update(&c, buffer, n); \
\
	if (!n) { \
		Final(binary, &c); \
		p = output; \
		for (i = 0; i < N; i++) { \
			*p++ = hex[(int)binary[i] >> 4]; \
			*p++ = hex[(int)binary[i] & 0x0f]; \
		} \
		*p = '\0'; \
	} else \
		output = NULL; \
\
	close(fd); \
	free(buffer); \
\
	return output; \
}

HASHFile(MD5File, MD5_CTX, MD5_Init, MD5_Update, MD5_Final, 16)
HASHFile(SHA1File, SHA_CTX, SHA1_Init, SHA1_Update, SHA1_Final, 20)
HASHFile(RMD160File, RIPEMD160_CTX,
	RIPEMD160_Init, RIPEMD160_Update, RIPEMD160_Final, 20)
